/*
** Original YM2413 emulator written by Mitsutaka Okazaki 2001
** Modified version by Rodrigo Cardoso to use in SegaX.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ym2413.h"
#include "def.h"

#define USE_SPEC_ENV_SPEED

struct _YM2413 YM2413;

#define PI 3.14159265358979
#define YM2413_TONE_NUM 2

extern const unsigned char default_inst[2][(16 + 3) * 16];

// Size of Sintable ( 1 -- 18 can be used, but 7 -- 14 recommended.)
#define PG_BITS 9
#define PG_WIDTH (1<<PG_BITS)

// Phase increment counter
#define DP_BITS 18
#define DP_WIDTH (1<<DP_BITS)
#define DP_BASE_BITS (DP_BITS - PG_BITS)

// Dynamic range (Accuracy of sin table)
#define DB_STEP (48.0/(1<<DB_BITS))
#define DB_BITS 8
#define DB_MUTE (1<<DB_BITS)

// Dynamic range of envelope
#define EG_STEP 0.375
#define EG_BITS 7
#define EG_MUTE (1<<EG_BITS)

// Dynamic range of total level
#define TL_STEP 0.75
#define TL_BITS 6
#define TL_MUTE (1<<TL_BITS)

// Dynamic range of sustine level
#define SL_STEP 3.0
#define SL_BITS 4
#define SL_MUTE (1<<SL_BITS)

#define EG2DB(d) ((d)*(long)(EG_STEP/DB_STEP))
#define TL2EG(d) ((d)*(long)(TL_STEP/EG_STEP))
#define SL2EG(d) ((d)*(long)(SL_STEP/EG_STEP))

#define DB_POS(x) (unsigned long)((x)/DB_STEP)
#define DB_NEG(x) (unsigned long)(DB_MUTE+DB_MUTE+(x)/DB_STEP)

// Bits for liner value
#define DB2LIN_AMP_BITS 11
#define SLOT_AMP_BITS (DB2LIN_AMP_BITS)

// Bits for envelope phase incremental counter  
#define EG_DP_BITS 22
#define EG_DP_WIDTH (1<<EG_DP_BITS)

// Bits for Pitch and Amp modulator
#define PM_PG_BITS 8
#define PM_PG_WIDTH (1<<PM_PG_BITS)
#define PM_DP_BITS 16
#define PM_DP_WIDTH (1<<PM_DP_BITS)
#define AM_PG_BITS 8
#define AM_PG_WIDTH (1<<AM_PG_BITS)
#define AM_DP_BITS 16
#define AM_DP_WIDTH (1<<AM_DP_BITS)

// PM table is calcurated by PM_AMP * pow(2,PM_DEPTH*sin(x)/1200)
#define PM_AMP_BITS 8
#define PM_AMP (1<<PM_AMP_BITS)

// PM speed(Hz) and depth(cent)
#define PM_SPEED 6.4
#define PM_DEPTH 13.75

// AM speed(Hz) and depth(dB)
#define AM_SPEED 3.7
#define AM_DEPTH 2.4

// Cut the lower b bit(s) off.
#define HIGHBITS(c,b) ((c)>>(b))

// Leave the lower b bit(s).
#define LOWBITS(c,b) ((c)&((1<<(b))-1))

// Expand x which is s bits to d bits.
#define EXPAND_BITS(x,s,d) ((x)<<((d)-(s)))

// Expand x which is s bits to d bits and fill expanded bits '1'
#define EXPAND_BITS_X(x,s,d) (((x)<<((d)-(s)))|((1<<((d)-(s)))-1))

// Adjust envelope speed which depends on sampling rate.
#define rate_adjust(x) (rate==49716?x:(unsigned long)((double)(x)*clk/72/rate + 0.5)) // +0.5 to round

#define MOD(x) ch[x].mod
#define CAR(x) ch[x].car

// Sampling rate
static unsigned long rate;
// Input clock
static unsigned long clk;

// WaveTable for each envelope amp
static unsigned short fullsintable[PG_WIDTH];
static unsigned short halfsintable[PG_WIDTH];

static unsigned long dphaseNoiseTable[512][8];

static unsigned short *waveform[2] = {fullsintable,halfsintable};

// LFO Table
static long pmtable[PM_PG_WIDTH];
static long amtable[AM_PG_WIDTH];

// Noise and LFO
static unsigned long pm_dphase;
static unsigned long am_dphase;

// dB to Liner table
static short DB2LIN_TABLE[(DB_MUTE + DB_MUTE)*2];

// Liner to Log curve conversion table (for Attack rate).
static unsigned short AR_ADJUST_TABLE[1<<EG_BITS];

// Empty voice data
static YM2413_PATCH null_patch = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

// Basic voice Data
static YM2413_PATCH default_patch[YM2413_TONE_NUM][(16 + 3) * 2]; 


// Definition of envelope mode
enum { SETTLE,ATTACK,DECAY,SUSHOLD,SUSTINE,RELEASE,FINISH };

// Phase incr table for Attack
static unsigned long dphaseARTable[16][16];
// Phase incr table for Decay and Release
static unsigned long dphaseDRTable[16][16];

// KSL + TL Table
static unsigned long tllTable[16][8][1<<TL_BITS][4];
static long rksTable[2][8][2];

// Phase incr table for PG
static unsigned long dphaseTable[512][8][16]; 

/* **************************************************
                  Create tables
*************************************************** */
INLINE static long Min(long i,long j)
{
  if(i < j) return i; else return j;
}

// Table for AR to LogCurve.
static void makeAdjustTable(void)
{
  long i;

  AR_ADJUST_TABLE[0] = (1 << EG_BITS);
  for(i = 1; i < 128; i++)
    AR_ADJUST_TABLE[i] = (unsigned short)((double)(1 << EG_BITS) - 1 - (1 << EG_BITS) * log(i) / log(128)); 
}


// Table for dB(0 -- (1<<DB_BITS)) to Liner(0 -- DB2LIN_AMP_WIDTH)
static void makeDB2LinTable(void)
{
  long i;

  for(i = 0; i < DB_MUTE + DB_MUTE; i++)
  {
    DB2LIN_TABLE[i] = (short)((double)((1 << DB2LIN_AMP_BITS) - 1) * pow(10, -(double)i * DB_STEP / 20));
    if(i >= DB_MUTE) DB2LIN_TABLE[i] = 0;
    DB2LIN_TABLE[i + DB_MUTE + DB_MUTE] = (short)(-DB2LIN_TABLE[i]);
  }
}

// Liner(+0.0 - +1.0) to dB((1<<DB_BITS) - 1 -- 0)
static long lin2db(double d)
{
  if(d == 0) return (DB_MUTE - 1);
  else return Min(-(long)(20.0*log10(d)/DB_STEP), DB_MUTE - 1); // 0 -- 127
}


// Sin Table
static void makeSinTable(void)
{
  long i;

  for(i = 0; i < PG_WIDTH / 4; i++)
    fullsintable[i] = (unsigned short)lin2db(sin(2.0*PI*i/PG_WIDTH));
  for(i = 0; i < PG_WIDTH / 4; i++)
    fullsintable[PG_WIDTH / 2 - 1 - i] = fullsintable[i];
  for(i = 0; i < PG_WIDTH / 2; i++)
    fullsintable[PG_WIDTH / 2 + i] = (unsigned short)(DB_MUTE + DB_MUTE + fullsintable[i]);
  for(i = 0; i < PG_WIDTH / 2; i++) halfsintable[i] = fullsintable[i];
  for(i = PG_WIDTH / 2; i < PG_WIDTH; i++) halfsintable[i] = fullsintable[0];
}

static void makeDphaseNoiseTable(void)
{
  int i, j;
  for(i = 0; i < 512; i++)
    for(j = 0;j < 8; j++) dphaseNoiseTable[i][j] = rate_adjust(i<<j);
}

// Table for Pitch Modulator
static void makePmTable(void)
{
  long i;
  for(i = 0; i < PM_PG_WIDTH; i++)
    pmtable[i] = (long)((double)PM_AMP * pow(2, (double)PM_DEPTH * sin(2.0 * PI * i / PM_PG_WIDTH) / 1200));
}

// Table for Amp Modulator
static void makeAmTable(void)
{
  long i;
  for(i = 0; i < AM_PG_WIDTH; i++)
    amtable[i] = (long)((double)AM_DEPTH / 2 / DB_STEP * (1.0 + sin(2.0 * PI * i / PM_PG_WIDTH)));
}

// Phase increment counter table 
static void makeDphaseTable(void)
{
  unsigned long fnum, block , ML;
  unsigned long mltable[16]= { 1, 1 * 2, 2 * 2, 3 * 2, 4 * 2, 5 * 2, 6 * 2, 7 * 2, 8 * 2,
                               9 * 2, 10 * 2, 10 * 2, 12 * 2, 12 * 2, 15 * 2, 15 * 2 };
  for(fnum = 0; fnum < 512; fnum++)
    for(block = 0; block < 8; block++)
      for(ML = 0; ML < 16; ML++)
        dphaseTable[fnum][block][ML] = rate_adjust(((fnum * mltable[ML]) << block) >> (20 - DP_BITS));
}

#define dB2(x) ((x)*2)

static void makeTllTable(void)
{
	static double kltable[16] = {
    dB2( 0.000), dB2( 9.000), dB2(12.000), dB2(13.875), dB2(15.000), dB2(16.125), dB2(16.875), dB2(17.625),
    dB2(18.000), dB2(18.750), dB2(19.125), dB2(19.500), dB2(19.875), dB2(20.250), dB2(20.625), dB2(21.000)
	};  
  long tmp;
  long fnum, block ,TL , KL;

  for(fnum = 0; fnum < 16; fnum++)
    for(block = 0; block < 8; block++)
      for(TL = 0; TL < 64; TL++)
        for(KL = 0; KL < 4; KL++)
        {
          if(KL==0) tllTable[fnum][block][TL][KL] = TL2EG(TL);
          else
          {
  	        tmp = (long)(kltable[fnum] - dB2(3.000) * (7 - block));
            if(tmp <= 0)
              tllTable[fnum][block][TL][KL] = TL2EG(TL);
            else 
              tllTable[fnum][block][TL][KL] = (unsigned long)((tmp >> (3 - KL)) / EG_STEP) + TL2EG(TL);
          }
       }
}

#ifdef USE_SPEC_ENV_SPEED
static double attacktime[16][4] = {
  {0,0,0,0},
  {1730.15, 1400.60, 1153.43, 988.66}, 
  {865.08, 700.30, 576.72, 494.33},
  {432.54, 350.15, 288.36, 247.16},
  {216.27, 175.07, 144.18, 123.58},
  {108.13, 87.54, 72.09, 61.79},
  {54.07, 43.77, 36.04, 30.90},
  {27.03, 21.88, 18.02, 15.45},
  {13.52, 10.94, 9.01, 7.72},
  {6.76, 5.47, 4.51, 3.86},
  {3.38, 2.74, 2.25, 1.93},
  {1.69, 1.37, 1.13, 0.97},
  {0.84, 0.70, 0.60, 0.54},
  {0.50, 0.42, 0.34, 0.30},
  {0.28, 0.22, 0.18, 0.14},
  {0.00, 0.00, 0.00, 0.00} 
};

static double decaytime[16][4] = {
  {0,0,0,0},
  {20926.60,16807.20,14006.00,12028.60},
  {10463.30,8403.58,7002.98,6014.32},
  {5231.64,4201.79,3501.49,3007.16},
  {2615.82,2100.89,1750.75,1503.58},
  {1307.91,1050.45,875.37,751.79},
  {653.95,525.22,437.69,375.90},
  {326.98,262.61,218.84,187.95},
  {163.49,131.31,109.42,93.97},
  {81.74,65.65,54.71,46.99},
  {40.87,32.83,27.36,23.49},
  {20.44,16.41,13.68,11.75},
  {10.22,8.21,6.84,5.87},
  {5.11,4.10,3.42,2.94},
  {2.55,2.05,1.71,1.47},
  {1.27,1.27,1.27,1.27} 
};
#endif

// Rate Table for Attack
static void makeDphaseARTable(void)
{
  long AR, Rks, RM, RL;
#ifdef USE_SPEC_ENV_SPEED
  unsigned long attacktable[16][4];

  for(RM = 0; RM < 16; RM++)
    for(RL = 0; RL < 4; RL++)
    {
      if     (RM == 0)  attacktable[RM][RL] = 0;
      else if(RM == 15) attacktable[RM][RL] = EG_DP_WIDTH;
      else              attacktable[RM][RL] = (unsigned long)((double)(1 << EG_DP_BITS)/(attacktime[RM][RL] * 3579545 / 72000));
    }
#endif

  for(AR = 0; AR < 16; AR++)
    for(Rks = 0; Rks < 16; Rks++)
    {
      RM = AR + (Rks>>2);
      RL = Rks&3;
      if(RM > 15) RM = 15;
      switch(AR)
      { 
        case 0:
          dphaseARTable[AR][Rks] = 0;
          break;
        case 15:
          dphaseARTable[AR][Rks] = EG_DP_WIDTH;
          break;
        default:
#ifdef USE_SPEC_ENV_SPEED
          dphaseARTable[AR][Rks] = rate_adjust(attacktable[RM][RL]);
#else
          dphaseARTable[AR][Rks] = rate_adjust((3 * (RL + 4) << (RM + 1)));
#endif
          break;
      }
    }
}

// Rate Table for Decay and Release
static void makeDphaseDRTable(void)
{
  long DR, Rks, RM, RL;

#ifdef USE_SPEC_ENV_SPEED
  unsigned long decaytable[16][4];

  for(RM = 0; RM < 16; RM++)
    for(RL = 0; RL < 4; RL++)
      if(RM == 0) decaytable[RM][RL] = 0;
      else        decaytable[RM][RL] = (unsigned long)((double)(1 << EG_DP_BITS) / (decaytime[RM][RL] * 3579545 / 72000));
#endif

  for(DR = 0; DR < 16; DR++)
    for(Rks = 0; Rks < 16; Rks++)
    {
      RM = DR + (Rks >> 2);
      RL = Rks & 3;
      if(RM > 15) RM = 15;
      switch(DR)
      { 
        case 0:
          dphaseDRTable[DR][Rks] = 0;
          break;
        default:
#ifdef USE_SPEC_ENV_SPEED
          dphaseDRTable[DR][Rks] = rate_adjust(decaytable[RM][RL]);
#else
          dphaseDRTable[DR][Rks] = rate_adjust((RL + 4) << (RM - 1));
#endif
          break;
      }
    }
}

static void makeRksTable(void)
{

  long fnum8, block, KR;

  for(fnum8 = 0; fnum8 < 2; fnum8++)
    for(block = 0; block < 8; block++)
      for(KR = 0; KR < 2; KR++)
      {
        if(KR != 0) rksTable[fnum8][block][KR] = (block << 1) + fnum8;
        else        rksTable[fnum8][block][KR] = block >> 1;
      }
}

void YM2413_Dump2Patch(const unsigned char *dump, YM2413_PATCH *patch)
{
  patch[0].AM = (dump[0] >> 7) & 1;
  patch[1].AM = (dump[1] >> 7) & 1;
  patch[0].PM = (dump[0] >> 6) & 1;
  patch[1].PM = (dump[1] >> 6) & 1;
  patch[0].EG = (dump[0] >> 5) & 1;
  patch[1].EG = (dump[1] >> 5) & 1;
  patch[0].KR = (dump[0] >> 4) & 1;
  patch[1].KR = (dump[1] >> 4) & 1;
  patch[0].ML = (dump[0]) & 15;
  patch[1].ML = (dump[1]) & 15;
  patch[0].KL = (dump[2] >> 6) & 3;
  patch[1].KL = (dump[3] >> 6) & 3;
  patch[0].TL = (dump[2]) & 63;
  patch[0].FB = (dump[3]) & 7;
  patch[0].WF = (dump[3] >> 3) & 1;
  patch[1].WF = (dump[3] >> 4) & 1;
  patch[0].AR = (dump[4] >> 4) & 15;
  patch[1].AR = (dump[5] >> 4) & 15;
  patch[0].DR = (dump[4]) & 15;
  patch[1].DR = (dump[5]) & 15;
  patch[0].SL = (dump[6] >> 4) & 15;
  patch[1].SL = (dump[7] >> 4) & 15;
  patch[0].RR = (dump[6]) & 15;
  patch[1].RR = (dump[7]) & 15;
}

void YM2413_GetDefaultPatch(long type, long num, YM2413_PATCH *patch)
{
  YM2413_Dump2Patch(default_inst[type]+num*16,patch);
}

static void makeDefaultPatch()
{
  long i, j;
  for(i = 0;i < YM2413_TONE_NUM; i++)
    for(j = 0;j < 19; j++) YM2413_GetDefaultPatch(i, j, &default_patch[i][j*2]);
}

void YM2413_SetPatch(const unsigned char *dump)
{
  YM2413_PATCH patch[2];
  int i;

  for(i = 0; i < 19; i++)
  {
    YM2413_Dump2Patch(dump + i * 16, patch);
    memcpy(&YM2413.patch[i * 2],     &patch[0], sizeof(YM2413_PATCH));
    memcpy(&YM2413.patch[i * 2 + 1], &patch[1], sizeof(YM2413_PATCH));
  }
}

void YM2413_Patch2Dump(const YM2413_PATCH *patch, unsigned char *dump)
{
  dump[0]  = (unsigned char)((patch[0].AM << 7) + (patch[0].PM << 6) + (patch[0].EG << 5) + (patch[0].KR << 4) + patch[0].ML);
  dump[1]  = (unsigned char)((patch[1].AM << 7) + (patch[1].PM << 6) + (patch[1].EG << 5) + (patch[1].KR << 4) + patch[1].ML);
  dump[2]  = (unsigned char)((patch[0].KL << 6) + (patch[0].TL));
  dump[3]  = (unsigned char)((patch[1].KL << 6) + (patch[1].WF << 4) + (patch[0].WF << 3) + patch[0].FB);
  dump[4]  = (unsigned char)((patch[0].AR << 4) + patch[0].DR);
  dump[5]  = (unsigned char)((patch[1].AR << 4) + patch[1].DR);
  dump[6]  = (unsigned char)((patch[0].SL << 4) + patch[0].RR);
  dump[7]  = (unsigned char)((patch[1].SL << 4) + patch[1].RR);
  dump[8]  = 0;
  dump[9]  = 0;
  dump[10] = 0;
  dump[11] = 0;
  dump[12] = 0;
  dump[13] = 0;
  dump[14] = 0;
  dump[15] = 0;
}

/* ***********************************************************
                      Calc Parameters
*********************************************************** */
INLINE static unsigned long calc_eg_dphase(YM2413_SLOT *slot)
{

  switch(slot->eg_mode)
  {
    case ATTACK:
      return dphaseARTable[slot->patch->AR][slot->rks];
    case DECAY:
      return dphaseDRTable[slot->patch->DR][slot->rks];
    case SUSHOLD:
      return 0;
    case SUSTINE:
      return dphaseDRTable[slot->patch->RR][slot->rks];
    case RELEASE:
      if     (slot->sustine)   return dphaseDRTable[5][slot->rks];
      else if(slot->patch->EG) return dphaseDRTable[slot->patch->RR][slot->rks];
      else                     return dphaseDRTable[7][slot->rks];
    case FINISH:
      return 0;
    default:
      return 0;
  }
}

/* ************************************************************
                    YM2413 internal interfaces
************************************************************ */
#define SLOT_BD1 12
#define SLOT_BD2 13
#define SLOT_HH 14
#define SLOT_SD 15
#define SLOT_TOM 16
#define SLOT_CYM 17

#define UPDATE_PG(S)  (S)->dphase = dphaseTable[(S)->fnum][(S)->block][(S)->patch->ML]
#define UPDATE_TLL(S)\
(((S)->type==0)?\
((S)->tll = tllTable[((S)->fnum)>>5][(S)->block][(S)->patch->TL][(S)->patch->KL]):\
((S)->tll = tllTable[((S)->fnum)>>5][(S)->block][(S)->volume][(S)->patch->KL]))
#define UPDATE_RKS(S) (S)->rks = rksTable[((S)->fnum)>>8][(S)->block][(S)->patch->KR] 
#define UPDATE_WF(S)  (S)->sintbl = waveform[(S)->patch->WF] 
#define UPDATE_EG(S)  (S)->eg_dphase = calc_eg_dphase(S) 
#define UPDATE_ALL(S)\
  UPDATE_PG(S);\
  UPDATE_TLL(S);\
  UPDATE_RKS(S);\
  UPDATE_WF(S); \
  UPDATE_EG(S) // EG should be updated last.

// Force Refresh (When external program changes some parameters).
void YM2413_ForceRefresh(void)
{
  long i;

  for(i=0; i<18;i++)
  {
    UPDATE_PG(YM2413.slot[i]);
    UPDATE_RKS(YM2413.slot[i]);
    UPDATE_TLL(YM2413.slot[i]);
    UPDATE_WF(YM2413.slot[i]);
    UPDATE_EG(YM2413.slot[i]);
  }
}

// Slot key on 
INLINE static void slotOn(YM2413_SLOT *slot)
{
  slot->eg_mode = ATTACK;
  slot->phase = 0;
  slot->eg_phase = 0;
}

// Slot key off
INLINE static void slotOff(YM2413_SLOT *slot)
{
  if(slot->eg_mode == ATTACK)
    slot->eg_phase = EXPAND_BITS(AR_ADJUST_TABLE[HIGHBITS(slot->eg_phase,EG_DP_BITS-EG_BITS)],EG_BITS,EG_DP_BITS);
  slot->eg_mode = RELEASE;
}

// Channel key on
INLINE static void keyOn(long i)
{
  if(!YM2413.slot_on_flag[i*2]) slotOn(&YM2413.MOD(i));
  if(!YM2413.slot_on_flag[i*2+1]) slotOn(&YM2413.CAR(i));
  YM2413.ch[i].key_status = 1;
}

// Channel key off
INLINE static void keyOff(long i)
{
  if(YM2413.slot_on_flag[i*2+1]) slotOff(&YM2413.CAR(i));
  YM2413.ch[i].key_status = 0;
}

INLINE static void keyOn_BD(void)  { keyOn(6); }
INLINE static void keyOn_SD(void)  { if(!YM2413.slot_on_flag[SLOT_SD])  slotOn(&YM2413.CAR(7)); }
INLINE static void keyOn_TOM(void) { if(!YM2413.slot_on_flag[SLOT_TOM]) slotOn(&YM2413.MOD(8)); }
INLINE static void keyOn_HH(void)  { if(!YM2413.slot_on_flag[SLOT_HH])  slotOn(&YM2413.MOD(7)); }
INLINE static void keyOn_CYM(void) { if(!YM2413.slot_on_flag[SLOT_CYM]) slotOn(&YM2413.CAR(8)); }

// Drum key off
INLINE static void keyOff_BD(void)  { keyOff(6); }
INLINE static void keyOff_SD(void)  { if(YM2413.slot_on_flag[SLOT_SD])  slotOff(&YM2413.CAR(7)); }
INLINE static void keyOff_TOM(void) { if(YM2413.slot_on_flag[SLOT_TOM]) slotOff(&YM2413.MOD(8)); }
INLINE static void keyOff_HH(void)  { if(YM2413.slot_on_flag[SLOT_HH])  slotOff(&YM2413.MOD(7)); }
INLINE static void keyOff_CYM(void) { if(YM2413.slot_on_flag[SLOT_CYM]) slotOff(&YM2413.CAR(8)); }

// Change a voice
INLINE static void SetPatch(long i, long num)
{
  YM2413.ch[i].patch_number = num;
  YM2413.MOD(i).patch = &YM2413.patch[num * 2 + 0];
  YM2413.CAR(i).patch = &YM2413.patch[num * 2 + 1];
}

// Change a rhythm voice
INLINE static void SetSlotPatch(YM2413_SLOT *slot, YM2413_PATCH *patch)
{
  slot->patch = patch;
}

// Set sustine parameter
INLINE static void SetSustine(long c, long sustine)
{
  YM2413.CAR(c).sustine = sustine;
  if(YM2413.MOD(c).type) YM2413.MOD(c).sustine = sustine;
}

// Volume : 6bit ( Volume register << 2 )
INLINE static void SetVolume(long c, long volume)
{
  YM2413.CAR(c).volume = volume;
}

INLINE static void SetSlotVolume(YM2413_SLOT *slot, long volume)
{
  slot->volume = volume;
}

// Set F-Number ( fnum : 9bit )
INLINE static void SetFNumber(long c, long fnum)
{
  YM2413.CAR(c).fnum = fnum;
  YM2413.MOD(c).fnum = fnum;
}

// Set Block data (block : 3bit )
INLINE static void SetBlock(long c, long block)
{
  YM2413.CAR(c).block = block;
  YM2413.MOD(c).block = block;
}

// Change Rhythm Mode
INLINE static void update_rythm_mode(void)
{
  if(YM2413.ch[6].patch_number & 0x10)
  {
    if(!(YM2413.slot_on_flag[SLOT_BD2] | (YM2413.reg[0x0E] & 32)))
    {
      YM2413.slot[SLOT_BD1]->eg_mode = FINISH;
      YM2413.slot[SLOT_BD2]->eg_mode = FINISH;
      SetPatch(6, YM2413.reg[0x36] >> 4);
    }
  }
  else if(YM2413.reg[0x0e]&32)
  {
    YM2413.ch[6].patch_number = 16;
    YM2413.slot[SLOT_BD1]->eg_mode = FINISH;
    YM2413.slot[SLOT_BD2]->eg_mode = FINISH;
    SetSlotPatch(YM2413.slot[SLOT_BD1], &YM2413.patch[16 * 2 + 0]);
    SetSlotPatch(YM2413.slot[SLOT_BD2], &YM2413.patch[16 * 2 + 1]);
  }

  if(YM2413.ch[7].patch_number & 0x10)
  {
    if(!((YM2413.slot_on_flag[SLOT_HH] && YM2413.slot_on_flag[SLOT_SD]) | (YM2413.reg[0x0E] & 32)))
    {
      YM2413.slot[SLOT_HH]->type = 0;
      YM2413.slot[SLOT_HH]->eg_mode = FINISH;
      YM2413.slot[SLOT_SD]->eg_mode = FINISH;
      SetPatch(7, YM2413.reg[0x37]>>4);
    }
  }
  else if(YM2413.reg[0x0e]&32)
  {
    YM2413.ch[7].patch_number = 17;
    YM2413.slot[SLOT_HH]->type = 1;
    YM2413.slot[SLOT_HH]->eg_mode = FINISH;
    YM2413.slot[SLOT_SD]->eg_mode = FINISH;
    SetSlotPatch(YM2413.slot[SLOT_HH], &YM2413.patch[17 * 2 + 0]);
    SetSlotPatch(YM2413.slot[SLOT_SD], &YM2413.patch[17 * 2 + 1]);
  }

  if(YM2413.ch[8].patch_number & 0x10)
  {
    if(!((YM2413.slot_on_flag[SLOT_CYM] && YM2413.slot_on_flag[SLOT_TOM]) | (YM2413.reg[0x0E] & 32)))
    {
      YM2413.slot[SLOT_TOM]->type = 0;
      YM2413.slot[SLOT_TOM]->eg_mode = FINISH;
      YM2413.slot[SLOT_CYM]->eg_mode = FINISH;
      SetPatch(8, YM2413.reg[0x38] >> 4);
    }
  }
  else if(YM2413.reg[0x0E] & 32)
  {
    YM2413.ch[8].patch_number = 18;
    YM2413.slot[SLOT_TOM]->type = 1;
    YM2413.slot[SLOT_TOM]->eg_mode = FINISH;
    YM2413.slot[SLOT_CYM]->eg_mode = FINISH;
    SetSlotPatch(YM2413.slot[SLOT_TOM], &YM2413.patch[18 * 2 + 0]);
    SetSlotPatch(YM2413.slot[SLOT_CYM], &YM2413.patch[18 * 2 + 1]);
  }
}

INLINE static void update_key_status(void)
{
  int ch;

  for(ch = 0; ch < 9; ch++)
    YM2413.slot_on_flag[ch*2] = YM2413.slot_on_flag[ch*2+1] = (YM2413.reg[0x20 + ch]) & 0x10;

  if(YM2413.reg[0x0E] & 32)
  {
    YM2413.slot_on_flag[SLOT_BD1] |= (YM2413.reg[0x0E] & 0x10);
    YM2413.slot_on_flag[SLOT_BD2] |= (YM2413.reg[0x0E] & 0x10);
    YM2413.slot_on_flag[SLOT_SD]  |= (YM2413.reg[0x0E] & 0x08);
    YM2413.slot_on_flag[SLOT_HH]  |= (YM2413.reg[0x0E] & 0x01);
    YM2413.slot_on_flag[SLOT_TOM] |= (YM2413.reg[0x0E] & 0x04);
    YM2413.slot_on_flag[SLOT_CYM] |= (YM2413.reg[0x0E] & 0x02);
  }
}

INLINE void YM2413_CopyPatch(long num, YM2413_PATCH *patch)
{
  memcpy(&YM2413.patch[num], patch, sizeof(YM2413_PATCH));
}

/* **********************************************************
                      Initializing
********************************************************** */
static void YM2413_SLOT_reset(YM2413_SLOT *slot)
{
  slot->sintbl    = waveform[0];
  slot->phase     = 0;
  slot->dphase    = 0;
  slot->output[0] = 0;
  slot->output[1] = 0;
  slot->feedback  = 0;
  slot->eg_mode   = SETTLE;
  slot->eg_phase  = EG_DP_WIDTH;
  slot->eg_dphase = 0;
  slot->rks       = 0;
  slot->tll       = 0;
  slot->sustine   = 0;
  slot->fnum      = 0;
  slot->block     = 0;
  slot->volume    = 0;
  slot->pgout     = 0;
  slot->egout     = 0;
  slot->patch     = &null_patch;
}

static void YM2413_CH_Reset(YM2413_CH *ch)
{
  YM2413_SLOT_reset(&ch->mod);
  YM2413_SLOT_reset(&ch->car);
  ch->mod.type   = 0;
  ch->car.type   = 1;
  ch->key_status = 0;
}

static void internal_refresh(void)
{
  makeDphaseTable();
  makeDphaseARTable();
  makeDphaseDRTable();
  makeDphaseNoiseTable();
  pm_dphase = (unsigned long)rate_adjust(PM_SPEED * PM_DP_WIDTH / (clk / 72));
  am_dphase = (unsigned long)rate_adjust(AM_SPEED * AM_DP_WIDTH / (clk / 72));
}

static void maketables(unsigned long c, unsigned long r)
{
  if(c != clk)
  {
    clk = c;
    makePmTable();
    makeAmTable();
    makeDB2LinTable(); 
    makeAdjustTable();
    makeTllTable();
    makeRksTable();
    makeSinTable();
    makeDefaultPatch();
  }

  if(r != rate)
  {
    rate = r;
    internal_refresh();
  }
}

void YM2413_Init(unsigned long clock, unsigned long srate)
{
  long i;

  maketables(clock, srate);

  for(i = 0; i < 9; i++)
  {
    YM2413.slot[i * 2 + 0] = &YM2413.ch[i].mod;
    YM2413.slot[i * 2 + 1] = &YM2413.ch[i].car;
    YM2413.ch[i].mod.type  = 0;
    YM2413.ch[i].car.type  = 1;
  }
  for(i = 0; i < 18; i++)
  {
    YM2413.slot[i]->plfo_am = &YM2413.lfo_am;
    YM2413.slot[i]->plfo_pm = &YM2413.lfo_pm;
  }
  YM2413.mask    = 0;
  YM2413.quality = 0;

  YM2413_Reset();
  YM2413_ResetPatch(0);
}

// Reset patch datas by system default.
void YM2413_ResetPatch(long type)
{
  long i;
  for(i = 0; i < 19 * 2; i++)
    YM2413_CopyPatch(i, &default_patch[type % YM2413_TONE_NUM][i]);
}

// Reset whole of YM2413 except patch datas.
void YM2413_Reset(void)
{
  long i;

  YM2413.adr = 0;
  YM2413.out = 0;
  YM2413.pm_phase = 0;
  YM2413.am_phase = 0;
  YM2413.noise_seed = 0xFFFF;
  YM2413.noiseA = 0;
  YM2413.noiseB = 0;
  YM2413.noiseA_phase = 0;
  YM2413.noiseB_phase = 0;
  YM2413.noiseA_dphase = 0;
  YM2413.noiseB_dphase = 0;
  YM2413.mask = 0;
  for(i = 0; i < 9; i++)
  {
    YM2413_CH_Reset(&YM2413.ch[i]);
    SetPatch(i, 0);
  }
  for(i = 0; i < 0x40; i++) YM2413_WriteReg(i, 0);
	YM2413.realstep = (unsigned long)((1 << 31) / rate);
  YM2413.opllstep = (unsigned long)((1 << 31) / (clk / 72));
	YM2413.oplltime = 0;
  for(i = 0; i < 14; i++) YM2413.pan[i] = 2;
  YM2413.sprev[0] = YM2413.sprev[1] = 0;
  YM2413.snext[0] = YM2413.snext[1] = 0;
}

void YM2413_SetRate(unsigned long r)
{
  if(YM2413.quality) rate = 49716; else rate = r;
  internal_refresh();
  rate = r;
}

void YM2413_SetQuality(unsigned long q)
{
  YM2413.quality = q;
  YM2413_SetRate(rate);
}

/* ********************************************************
                 Generate wave data
******************************************************** */
// Convert Amp(0 to EG_HEIGHT) to Phase(0 to 2PI).
#if ( SLOT_AMP_BITS - PG_BITS ) > 0
#define wave2_2pi(e)  ( (e) >> ( SLOT_AMP_BITS - PG_BITS ))
#else
#define wave2_2pi(e)  ( (e) << ( PG_BITS - SLOT_AMP_BITS ))
#endif

// Convert Amp(0 to EG_HEIGHT) to Phase(0 to 4PI).
#if ( SLOT_AMP_BITS - PG_BITS - 1 ) == 0
#define wave2_4pi(e)  (e)
#elif ( SLOT_AMP_BITS - PG_BITS - 1 ) > 0
#define wave2_4pi(e)  ( (e) >> ( SLOT_AMP_BITS - PG_BITS - 1 ))
#else
#define wave2_4pi(e)  ( (e) << ( 1 + PG_BITS - SLOT_AMP_BITS ))
#endif

// Convert Amp(0 to EG_HEIGHT) to Phase(0 to 8PI).
#if ( SLOT_AMP_BITS - PG_BITS - 2 ) == 0
#define wave2_8pi(e)  (e)
#elif ( SLOT_AMP_BITS - PG_BITS - 2 ) > 0
#define wave2_8pi(e)  ( (e) >> ( SLOT_AMP_BITS - PG_BITS - 2 ))
#else
#define wave2_8pi(e)  ( (e) << ( 2 + PG_BITS - SLOT_AMP_BITS ))
#endif

// Update Noise unit
static void update_noise(void)
{
  if(YM2413.noise_seed & 1) YM2413.noise_seed ^= 0x24000;
  YM2413.noise_seed >>= 1;
  YM2413.whitenoise  = YM2413.noise_seed & 1 ? DB_POS(6.0) : DB_NEG(6.0);

  YM2413.noiseA_phase += YM2413.noiseA_dphase;
  YM2413.noiseB_phase += YM2413.noiseB_dphase;

  YM2413.noiseA_phase &= (0x40 << 11) - 1;
  YM2413.noiseA = YM2413.noiseA_phase & (0x03 << 11) ? DB_POS(3.0) : DB_NEG(3.0);

  YM2413.noiseB_phase &= (0x10 << 11) - 1;
  YM2413.noiseB = YM2413.noiseB_phase & (0x0A << 11) ? DB_POS(3.0) : DB_NEG(3.0);
}

// Update AM, PM unit
static void update_ampm(void)
{
  YM2413.pm_phase = (YM2413.pm_phase + pm_dphase)&(PM_DP_WIDTH - 1);
  YM2413.am_phase = (YM2413.am_phase + am_dphase)&(AM_DP_WIDTH - 1);
  YM2413.lfo_am = amtable[HIGHBITS(YM2413.am_phase, AM_DP_BITS - AM_PG_BITS)];
  YM2413.lfo_pm = pmtable[HIGHBITS(YM2413.pm_phase, PM_DP_BITS - PM_PG_BITS)];
}

// PG
INLINE static unsigned long calc_phase(YM2413_SLOT *slot)
{
  if(slot->patch->PM)
    slot->phase += (slot->dphase * (*(slot->plfo_pm))) >> PM_AMP_BITS;
  else
    slot->phase += slot->dphase;

  slot->phase &= (DP_WIDTH - 1);

  return HIGHBITS(slot->phase, DP_BASE_BITS);
}

// EG
static unsigned long calc_envelope(YM2413_SLOT *slot)
{
  #define S2E(x) (SL2EG((long)(x/SL_STEP))<<(EG_DP_BITS-EG_BITS)) 

  static unsigned long SL[16] = 
  {
    S2E( 0.0), S2E( 3.0), S2E( 6.0), S2E( 9.0), S2E(12.0), S2E(15.0), S2E(18.0), S2E(21.0),
    S2E(24.0), S2E(27.0), S2E(30.0), S2E(33.0), S2E(36.0), S2E(39.0), S2E(42.0), S2E(48.0)
  };

  unsigned long egout;

  switch(slot->eg_mode)
  {

    case ATTACK:
      egout = AR_ADJUST_TABLE[HIGHBITS(slot->eg_phase, EG_DP_BITS - EG_BITS)];
      slot->eg_phase += slot->eg_dphase;
      if(EG_DP_WIDTH & slot->eg_phase)
      {
        egout = 0;
        slot->eg_phase= 0;
        slot->eg_mode = DECAY;
        UPDATE_EG(slot);
      }
      break;

    case DECAY:
      egout = HIGHBITS(slot->eg_phase, EG_DP_BITS - EG_BITS);
      slot->eg_phase += slot->eg_dphase;
      if(slot->eg_phase >= SL[slot->patch->SL])
      {
        if(slot->patch->EG)
        {
          slot->eg_phase = SL[slot->patch->SL];
	        slot->eg_mode = SUSHOLD;
          UPDATE_EG(slot);
        }
        else
        {
          slot->eg_phase = SL[slot->patch->SL];
          slot->eg_mode = SUSTINE;
          UPDATE_EG(slot);
        }
      }
      break;

    case SUSHOLD:
      egout = HIGHBITS(slot->eg_phase, EG_DP_BITS - EG_BITS);
      if(slot->patch->EG == 0)
      {
        slot->eg_mode = SUSTINE;
        UPDATE_EG(slot);
      }
      break;

    case SUSTINE:
    case RELEASE:
      egout = HIGHBITS(slot->eg_phase, EG_DP_BITS - EG_BITS); 
      slot->eg_phase += slot->eg_dphase;
      if(egout >= (1<<EG_BITS))
      {
        slot->eg_mode = FINISH;
        egout = (1<<EG_BITS) - 1;
      }
      break;

    case FINISH:
      egout = (1<<EG_BITS) - 1;
      break;

    default:
      egout = (1<<EG_BITS) - 1;
      break;
  }

  if(slot->patch->AM) egout = EG2DB(egout+slot->tll) + *(slot->plfo_am);
  else egout = EG2DB(egout+slot->tll) ;

  if(egout >= DB_MUTE) egout = DB_MUTE-1;
  return egout;

}

// CARRIOR
INLINE static long calc_slot_car(YM2413_SLOT *slot, long fm)
{
  slot->egout = calc_envelope(slot); 
  slot->pgout = calc_phase(slot);
  slot->output[1] = slot->output[0];

  if(slot->egout>=(DB_MUTE-1))
  {
    slot->output[0] = 0;
  }
  else
  {
    slot->output[0] = DB2LIN_TABLE[slot->sintbl[(slot->pgout+wave2_8pi(fm))&(PG_WIDTH-1)] + slot->egout];
  }

  return (slot->output[1] + slot->output[0]) >> 1;
}

// MODULATOR
INLINE static long calc_slot_mod(YM2413_SLOT *slot)
{
  long fm;

  slot->output[1] = slot->output[0];
  slot->egout = calc_envelope(slot); 
  slot->pgout = calc_phase(slot);

  if(slot->egout>=(DB_MUTE-1))
  {
    slot->output[0] = 0;
  }
  else if(slot->patch->FB!=0)
  {
    fm = wave2_4pi(slot->feedback) >> (7 - slot->patch->FB);
    slot->output[0] = DB2LIN_TABLE[slot->sintbl[(slot->pgout+fm)&(PG_WIDTH-1)] + slot->egout];
  }
  else
  {
    slot->output[0] = DB2LIN_TABLE[slot->sintbl[slot->pgout] + slot->egout];
  }

  slot->feedback = (slot->output[1] + slot->output[0]) >> 1;

  return slot->feedback;

}

// TOM
INLINE static long calc_slot_tom(YM2413_SLOT *slot)
{

  slot->egout = calc_envelope(slot); 
  slot->pgout = calc_phase(slot);
  if(slot->egout>=(DB_MUTE-1)) return 0;
  
  return DB2LIN_TABLE[slot->sintbl[slot->pgout] + slot->egout];

}

// SNARE
INLINE static long calc_slot_snare(YM2413_SLOT *slot, unsigned long whitenoise)
{
  slot->egout = calc_envelope(slot); 
  slot->pgout = calc_phase(slot);
  if(slot->egout>=(DB_MUTE-1)) return 0;

  if(slot->pgout & (1<<(PG_BITS-1)))
    return (DB2LIN_TABLE[slot->egout] + DB2LIN_TABLE[slot->egout+whitenoise]) >> 1;
  else
    return (DB2LIN_TABLE[DB_MUTE + DB_MUTE + slot->egout] + DB2LIN_TABLE[slot->egout+whitenoise]) >> 1;
}

// TOP-CYM
INLINE static long calc_slot_cym(YM2413_SLOT *slot, long a, long b)
{
  slot->egout = calc_envelope(slot);
  if(slot->egout>=(DB_MUTE-1))
  {
    return 0;
  }
  else
  {
    return (DB2LIN_TABLE[slot->egout+a] + DB2LIN_TABLE[slot->egout+b]) >> 1;
  }
}

// HI-HAT
INLINE static long calc_slot_hat(YM2413_SLOT *slot, long a, long b, unsigned long whitenoise)
{
  slot->egout = calc_envelope(slot);
  if(slot->egout >= (DB_MUTE - 1)) return 0;
  else return (DB2LIN_TABLE[slot->egout + whitenoise] >> 2) + (DB2LIN_TABLE[slot->egout + a] >> 3) + (DB2LIN_TABLE[slot->egout + b] >> 3);
}

INLINE short calc(void)
{
  long inst = 0 , perc = 0;
  long i;

  update_ampm();
  update_noise();

  for(i = 0; i < 6; i++)
    if(!(YM2413.mask & YM2413_MASK_CH(i)) && (YM2413.CAR(i).eg_mode != FINISH))
      inst += calc_slot_car(&YM2413.CAR(i), calc_slot_mod(&YM2413.MOD(i)));

  // CH6
  if(YM2413.ch[6].patch_number <= 15)
  {
    if(!(YM2413.mask & YM2413_MASK_CH(6)) && (YM2413.CAR(6).eg_mode != FINISH))
      inst += calc_slot_car(&YM2413.CAR(6), calc_slot_mod(&YM2413.MOD(6)));
  }
  else
  {
    if(!(YM2413.mask & YM2413_MASK_BD) && (YM2413.CAR(6).eg_mode != FINISH))
      perc += calc_slot_car(&YM2413.CAR(6), calc_slot_mod(&YM2413.MOD(6)));
  }

  // CH7
  if(YM2413.ch[7].patch_number <= 15)
  {
    if(!(YM2413.mask & YM2413_MASK_CH(7)) && (YM2413.CAR(7).eg_mode != FINISH))
      inst += calc_slot_car(&YM2413.CAR(7), calc_slot_mod(&YM2413.MOD(7)));
  }
  else
  {
    if(!(YM2413.mask & YM2413_MASK_HH) && (YM2413.MOD(7).eg_mode != FINISH))
       perc += calc_slot_hat(&YM2413.MOD(7), YM2413.noiseA, YM2413.noiseB, YM2413.whitenoise);
    if(!(YM2413.mask & YM2413_MASK_SD) && (YM2413.CAR(7).eg_mode != FINISH))
        perc += calc_slot_snare(&YM2413.CAR(7), YM2413.whitenoise);
  }
  
  // CH8
  if(YM2413.ch[8].patch_number <= 15)
  {
    if(!(YM2413.mask & YM2413_MASK_CH(8)) && (YM2413.CAR(8).eg_mode != FINISH))
      inst += calc_slot_car(&YM2413.CAR(8), calc_slot_mod(&YM2413.MOD(8)));
  }
  else
  {
    if(!(YM2413.mask & YM2413_MASK_TOM) && (YM2413.MOD(8).eg_mode != FINISH))
       perc += calc_slot_tom(&YM2413.MOD(8));
    if(!(YM2413.mask & YM2413_MASK_CYM) && (YM2413.CAR(8).eg_mode != FINISH))
       perc += calc_slot_cym(&YM2413.CAR(8), YM2413.noiseA, YM2413.noiseB);
  }

  return (short)(inst + (perc << 1));
}

void YM2413_Update(int **buf, int len)
{
  int n;

  if(YM2413.quality)
  {
    for(n = 0; n < len; buf[0]++, buf[1]++, n++)
    {
      while(YM2413.realstep > YM2413.oplltime)
      {
        YM2413.oplltime += YM2413.opllstep;
        YM2413.prev      = YM2413.next;
        YM2413.next      = calc();
      }
      YM2413.oplltime -= YM2413.realstep;
      YM2413.out = (short)(((double)YM2413.next * (YM2413.opllstep - YM2413.oplltime)
                 + (double)YM2413.prev * YM2413.oplltime) / YM2413.opllstep);

      *buf[0] = *buf[1] = (short)YM2413.out;
    }
  }
  else
  {
    for(n = 0; n < len; buf[0]++, buf[1]++, n++) *buf[0] = *buf[1] = calc();
  }
}

unsigned long YM2413_SetMask(unsigned long mask)
{
  unsigned long ret;
  
  ret = YM2413.mask;
  YM2413.mask = mask;
  return ret;
}

unsigned long YM2413_ToggleMask(unsigned long mask)
{
  unsigned long ret;
  
  ret = YM2413.mask;
  YM2413.mask ^= mask;
  return ret;
}

/* ***************************************************
                       I/O Ctrl
**************************************************** */
void YM2413_WriteReg(unsigned long reg, unsigned long data)
{
  long i,v,ch;

  data = data & 0xFF;
  reg  = reg & 0x3F;
  YM2413.reg[reg] = (unsigned char)data;

  switch(reg)
  {
    case 0x00:
      YM2413.patch[0].AM = (data >> 7) & 1;
      YM2413.patch[0].PM = (data >> 6) & 1;
      YM2413.patch[0].EG = (data >> 5) & 1;
      YM2413.patch[0].KR = (data >> 4) & 1;
      YM2413.patch[0].ML = (data) & 15;
      for(i = 0; i < 9; i++)
      {
        if(YM2413.ch[i].patch_number == 0)
        {
          UPDATE_PG(&YM2413.MOD(i));
          UPDATE_RKS(&YM2413.MOD(i));
          UPDATE_EG(&YM2413.MOD(i));
        }
      }
      break;

    case 0x01:
      YM2413.patch[1].AM = (data >> 7) & 1;
      YM2413.patch[1].PM = (data >> 6) & 1;
      YM2413.patch[1].EG = (data >> 5) & 1;
      YM2413.patch[1].KR = (data >> 4) & 1;
      YM2413.patch[1].ML = (data) & 15;
      for(i = 0; i < 9; i++)
      {
        if(YM2413.ch[i].patch_number == 0)
        {
          UPDATE_PG(&YM2413.CAR(i));
          UPDATE_RKS(&YM2413.CAR(i));
          UPDATE_EG(&YM2413.CAR(i));
        }
      }
      break;

    case 0x02:
      YM2413.patch[0].KL = (data >> 6) & 3;
      YM2413.patch[0].TL = (data) & 63;
      for(i = 0; i < 9; i++)
      {
        if(YM2413.ch[i].patch_number == 0)
        {
          UPDATE_TLL(&YM2413.MOD(i));
        }
      }
      break;

    case 0x03:
      YM2413.patch[1].KL = (data >> 6) & 3;
      YM2413.patch[1].WF = (data >> 4) & 1;
      YM2413.patch[0].WF = (data >> 3) & 1;
      YM2413.patch[0].FB = (data) & 7;
      for(i = 0; i < 9; i++)
      {
        if(YM2413.ch[i].patch_number == 0)
        {
          UPDATE_WF(&YM2413.MOD(i));
          UPDATE_WF(&YM2413.CAR(i));
        }
      }
      break;

    case 0x04:
      YM2413.patch[0].AR = (data >> 4) & 15;
      YM2413.patch[0].DR = (data) & 15;
      for(i = 0; i < 9; i++)
      {
        if(YM2413.ch[i].patch_number == 0)
        {
          UPDATE_EG(&YM2413.MOD(i));
        }
      }
      break;

    case 0x05:
      YM2413.patch[1].AR = (data >> 4) & 15;
      YM2413.patch[1].DR = (data) & 15;
      for(i = 0; i < 9; i++)
      {
        if(YM2413.ch[i].patch_number == 0)
        {
          UPDATE_EG(&YM2413.CAR(i));
        }
      }
      break;

    case 0x06:
      YM2413.patch[0].SL = (data >> 4) & 15;
      YM2413.patch[0].RR = (data) & 15;
      for(i = 0; i < 9; i++)
      {
        if(YM2413.ch[i].patch_number == 0)
        {
          UPDATE_EG(&YM2413.MOD(i));
        }
      }
      break;

    case 0x07:
      YM2413.patch[1].SL = (data >> 4) & 15;
      YM2413.patch[1].RR = (data) & 15;
      for(i = 0; i < 9; i++)
      {
        if(YM2413.ch[i].patch_number == 0)
        {
          UPDATE_EG(&YM2413.CAR(i));
        }
      }
      break;

    case 0x0E:
      update_rythm_mode();
      if(data & 32)
      {
        if(data & 0x10) keyOn_BD();  else keyOff_BD();
        if(data & 0x08) keyOn_SD();  else keyOff_SD();
        if(data & 0x04) keyOn_TOM(); else keyOff_TOM();
        if(data & 0x02) keyOn_CYM(); else keyOff_CYM();
        if(data & 0x01) keyOn_HH();  else keyOff_HH();
      }
      update_key_status();

      UPDATE_ALL(&YM2413.MOD(6));
      UPDATE_ALL(&YM2413.CAR(6));
      UPDATE_ALL(&YM2413.MOD(7));  
      UPDATE_ALL(&YM2413.CAR(7));
      UPDATE_ALL(&YM2413.MOD(8));
      UPDATE_ALL(&YM2413.CAR(8));        

      break;

    case 0x0F:
      break;

    case 0x10:  case 0x11:  case 0x12:  case 0x13:
    case 0x14:  case 0x15:  case 0x16:  case 0x17:
    case 0x18:
      ch = reg - 0x10;
      SetFNumber(ch, data + ((YM2413.reg[0x20 + ch] & 1) << 8));
      UPDATE_ALL(&YM2413.MOD(ch));
      UPDATE_ALL(&YM2413.CAR(ch));
      switch(reg)
      {
      case 0x17:
        YM2413.noiseA_dphase = dphaseNoiseTable[data + ((YM2413.reg[0x27]&1)<<8)][(YM2413.reg[0x27]>>1)&7];
        break;
      case 0x18:
        YM2413.noiseB_dphase = dphaseNoiseTable[data + ((YM2413.reg[0x28]&1)<<8)][(YM2413.reg[0x28]>>1)&7];
        break;
      default:
        break;
      }
      break;

    case 0x20:  case 0x21:  case 0x22:  case 0x23:
    case 0x24:  case 0x25:  case 0x26:  case 0x27:
    case 0x28:
      ch = reg - 0x20;
      SetFNumber(ch, ((data&1)<<8) + YM2413.reg[0x10+ch]);
      SetBlock(ch, (data>>1)&7 );
      SetSustine(ch, (data>>5)&1);
      if(data & 0x10) keyOn(ch); else keyOff(ch);
      UPDATE_ALL(&YM2413.MOD(ch));
      UPDATE_ALL(&YM2413.CAR(ch));
      switch(reg)
      {
      case 0x27:
        YM2413.noiseA_dphase = dphaseNoiseTable[((data&1)<<8) + YM2413.reg[0x17]][(data>>1)&7];
        break;
      case 0x28:
        YM2413.noiseB_dphase = dphaseNoiseTable[((data&1)<<8) + YM2413.reg[0x18]][(data>>1)&7];
        break;
      default:
        break;
      }
      update_key_status();
      update_rythm_mode();
      break;

    case 0x30: case 0x31: case 0x32: case 0x33: case 0x34:
    case 0x35: case 0x36: case 0x37: case 0x38:
      i = (data >> 4) & 15;
      v = data & 15;
      if((YM2413.reg[0x0E] & 32) && (reg >= 0x36))
      {
        switch(reg)
        {
         case 0x37 :
           SetSlotVolume(&YM2413.MOD(7), i << 2);
           break;
         case 0x38 :
           SetSlotVolume(&YM2413.MOD(8), i << 2);
           break;
         default:
           break;
        }
      }
      else
      { 
        SetPatch(reg - 0x30, i);
      }
      SetVolume(reg - 0x30, v << 2);
      UPDATE_ALL(&YM2413.MOD(reg - 0x30));
      UPDATE_ALL(&YM2413.CAR(reg - 0x30));
      break;

    default:
      break;

  }

}

void YM2413_WriteIO(unsigned long adr, unsigned long val)
{
  if(adr & 1) YM2413_WriteReg(YM2413.adr, val);
  else YM2413.adr = val;
}

// STEREO MODE (OPT)
void YM2413_SetPan(unsigned long ch, unsigned long pan)
{
  YM2413.pan[ch&15] = pan&3;
}

static void calc_stereo(long out[2])
{
  long b[4] = { 0,0,0,0 }; // Ignore, Right, Left, Center
  long r[4] = { 0,0,0,0 }; // Ignore, Right, Left, Center
  long i;

  update_ampm();
  update_noise();

  for(i = 0; i < 6; i++)
    if(!(YM2413.mask & YM2413_MASK_CH(i)) && (YM2413.CAR(i).eg_mode != FINISH))
       b[YM2413.pan[i]] += calc_slot_car(&YM2413.CAR(i), calc_slot_mod(&YM2413.MOD(i)));


  if(YM2413.ch[6].patch_number <= 15)
  {
    if(!(YM2413.mask & YM2413_MASK_CH(6)) && (YM2413.CAR(6).eg_mode != FINISH))
      b[YM2413.pan[6]] += calc_slot_car(&YM2413.CAR(6), calc_slot_mod(&YM2413.MOD(6)));
  }
  else
  {
    if(!(YM2413.mask & YM2413_MASK_BD) && (YM2413.CAR(6).eg_mode != FINISH))
      r[YM2413.pan[9]] += calc_slot_car(&YM2413.CAR(6), calc_slot_mod(&YM2413.MOD(6)));
  }

  if(YM2413.ch[7].patch_number <= 15)
  {
    if(!(YM2413.mask & YM2413_MASK_CH(7)) && (YM2413.CAR(7).eg_mode != FINISH))
      b[YM2413.pan[7]] += calc_slot_car(&YM2413.CAR(7), calc_slot_mod(&YM2413.MOD(7)));
  }
  else
  {
    if(!(YM2413.mask & YM2413_MASK_HH) && (YM2413.MOD(7).eg_mode != FINISH))
      r[YM2413.pan[10]] += calc_slot_hat(&YM2413.MOD(7), YM2413.noiseA, YM2413.noiseB, YM2413.whitenoise);
    if(!(YM2413.mask & YM2413_MASK_SD) && (YM2413.CAR(7).eg_mode != FINISH))
      r[YM2413.pan[11]] += calc_slot_snare(&YM2413.CAR(7), YM2413.whitenoise);
  }

  if(YM2413.ch[8].patch_number <= 15)
  {
    if(!(YM2413.mask & YM2413_MASK_CH(8)) && (YM2413.CAR(8).eg_mode != FINISH))
      b[YM2413.pan[8]] += calc_slot_car(&YM2413.CAR(8), calc_slot_mod(&YM2413.MOD(8)));
  }
  else
  {
    if(!(YM2413.mask & YM2413_MASK_TOM) && (YM2413.MOD(8).eg_mode != FINISH))
      r[YM2413.pan[12]] += calc_slot_tom(&YM2413.MOD(8));
    if(!(YM2413.mask & YM2413_MASK_CYM) && (YM2413.CAR(8).eg_mode != FINISH))
      r[YM2413.pan[13]] += calc_slot_cym(&YM2413.CAR(8), YM2413.noiseA, YM2413.noiseB);
  }

  out[1] = b[1] + b[3] + ((r[1] + r[3])<< 1);
  out[0] = b[2] + b[3] + ((r[2] + r[3])<< 1);
}

void YM2413_CalcStereo(long out[2])
{
  if(!YM2413.quality) 
  {
    calc_stereo(out);
    return;
  }

  while (YM2413.realstep > YM2413.oplltime)
  {
    YM2413.oplltime += YM2413.opllstep;
    YM2413.sprev[0] = YM2413.snext[0];
    YM2413.sprev[1] = YM2413.snext[1];
    calc_stereo(YM2413.snext);
  }

  YM2413.oplltime -= YM2413.realstep;
  out[0] = (short)(((double)YM2413.snext[0] * (YM2413.opllstep - YM2413.oplltime)
              +(double)YM2413.sprev[0] * YM2413.oplltime)/YM2413.opllstep);
  out[1] = (short)(((double)YM2413.snext[1] * (YM2413.opllstep - YM2413.oplltime)
              +(double)YM2413.sprev[1] * YM2413.oplltime)/YM2413.opllstep);
}

const unsigned char default_inst[2][(16 + 3) * 16] = {
{ 
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x61,0x61,0x1e,0x17,0xf0,0x7f,0x07,0x17,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x13,0x41,0x0f,0x0d,0xce,0xf5,0x43,0x23,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x03,0x01,0x9a,0x04,0xf3,0xf4,0x13,0x23,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x21,0x61,0x1d,0x07,0xfa,0x64,0x30,0x28,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x22,0x21,0x1e,0x06,0xf0,0x76,0x18,0x28,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x31,0x02,0x16,0x05,0x90,0x71,0x00,0x10,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x21,0x61,0x1d,0x07,0x82,0x80,0x10,0x17,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x23,0x21,0x2d,0x16,0xc0,0x70,0x07,0x07,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x61,0x21,0x1b,0x06,0x64,0x65,0x18,0x18,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x61,0x61,0x0c,0x18,0x85,0xa0,0x79,0x07,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x23,0x21,0x87,0x11,0xf0,0xa4,0x00,0xf7,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x97,0xe1,0x28,0x07,0xff,0xf3,0x02,0xf8,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x61,0x10,0x0c,0x05,0xf2,0xc4,0x40,0xc8,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x01,0x01,0x56,0x03,0xb4,0xb2,0x23,0x58,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x61,0x41,0x89,0x03,0xf1,0xf4,0xf0,0x13,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x07,0x21,0x14,0x00,0xee,0xf8,0xff,0xf8,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x23,0x22,0x00,0x00,0xd8,0xf8,0xf8,0xf7,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x25,0x18,0x00,0x00,0xf8,0xda,0xf8,0x55,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
},
{ 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x33, 0x01, 0x09, 0x0e, 0x94, 0x90, 0x40, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x13, 0x41, 0x0f, 0x0d, 0xce, 0xd3, 0x43, 0x13, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x01, 0x12, 0x1b, 0x06, 0xff, 0xd2, 0x00, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x61, 0x61, 0x1b, 0x07, 0xaf, 0x63, 0x20, 0x28, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x22, 0x21, 0x1e, 0x06, 0xf0, 0x76, 0x08, 0x28, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x66, 0x21, 0x15, 0x00, 0x93, 0x94, 0x20, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x21, 0x61, 0x1c, 0x07, 0x82, 0x81, 0x10, 0x17, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x23, 0x21, 0x20, 0x1f, 0xc0, 0x71, 0x07, 0x47, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x25, 0x31, 0x26, 0x05, 0x64, 0x41, 0x18, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x17, 0x21, 0x28, 0x07, 0xff, 0x83, 0x02, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x97, 0x81, 0x25, 0x07, 0xcf, 0xc8, 0x02, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x21, 0x21, 0x54, 0x0f, 0x80, 0x7f, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x01, 0x01, 0x56, 0x03, 0xd3, 0xb2, 0x43, 0x58, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x31, 0x21, 0x0c, 0x03, 0x82, 0xc0, 0x40, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x21, 0x01, 0x0c, 0x03, 0xd4, 0xd3, 0x40, 0x84, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x04, 0x21, 0x28, 0x00, 0xdf, 0xf8, 0xff, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x23, 0x22, 0x00, 0x00, 0xa8, 0xf8, 0xf8, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x25, 0x18, 0x00, 0x00, 0xf8, 0xa9, 0xf8, 0x55, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
} };
