/*
** Original YM2413 emulator written by Mitsutaka Okazaki 2001
** Modified version by Rodrigo Cardoso to use in SegaX.
*/

#ifndef _YM2413_H
#define _YM2413_H

#include "def.h"

#ifdef __cplusplus
extern "C" {
#endif

enum {
  YM2413_2413_TONE,
  YM2413_VRC7_TONE
};

// voice data
typedef struct _YM2413_PATCH {
  unsigned long TL, FB, EG, ML, AR, DR, SL, RR, KR, KL, AM, PM, WF;
} YM2413_PATCH;

// slot
typedef struct _YM2413_SLOT {
  YM2413_PATCH   *patch;
  long            type;      // 0 : modulator 1 : carrier
  // OUTPUT
  long            feedback;
  long            output[5]; // Output value of slot
  // for Phase Generator (PG)
  unsigned short *sintbl;    // Wavetable
  unsigned long   phase;     // Phase
  unsigned long   dphase;    // Phase increment amount
  unsigned long   pgout;     // output
  // for Envelope Generator (EG)
  long            fnum;      // F-Number
  long            block;     // Block
  long            volume;    // Current volume
  long            sustine;   // Sustine 1 = ON, 0 = OFF
  unsigned long   tll;	     // Total Level + Key scale level*/
  unsigned long   rks;       // Key scale offset (Rks)
  long            eg_mode;   // Current state
  unsigned long   eg_phase;  // Phase
  unsigned long   eg_dphase; // Phase increment amount
  unsigned long   egout;     // output
  // refer to opll->
  long           *plfo_pm;
  long           *plfo_am;
} YM2413_SLOT;

// Channel
typedef struct _YM2413_CH {
  long          patch_number;
  long          key_status;
  YM2413_SLOT   mod, car;
} YM2413_CH;

// Mask
#define YM2413_MASK_CH(x) (1 << (x))
#define YM2413_MASK_HH    (1 << (9))
#define YM2413_MASK_CYM   (1 << (10))
#define YM2413_MASK_TOM   (1 << (11))
#define YM2413_MASK_SD    (1 << (12))
#define YM2413_MASK_BD    (1 << (13))
#define YM2413_MASK_RYTHM (YM2413_MASK_HH | YM2413_MASK_CYM | YM2413_MASK_TOM | YM2413_MASK_SD | YM2413_MASK_BD)

// opll
struct _YM2413 {
  // Register
  unsigned char reg[0x40];
  unsigned long realstep;
  unsigned long oplltime;
  unsigned long opllstep;
  unsigned long adr;
  long          out, prev, next;
  long          sprev[2],snext[2];
  unsigned long pan[16];
  long          slot_on_flag[18];
  // Pitch Modulator
  unsigned long pm_phase;
  long          lfo_pm;
  // Amp Modulator
  long          am_phase;
  long          lfo_am;
  unsigned long quality;
  // Noise Generator
  unsigned long noise_seed;
  unsigned long whitenoise;
  unsigned long noiseA;
  unsigned long noiseB;
  unsigned long noiseA_phase;
  unsigned long noiseB_phase;
  unsigned long noiseA_dphase;
  unsigned long noiseB_dphase;
  // Channel & Slot
  YM2413_CH     ch[9];
  YM2413_SLOT  *slot[18];
  // Voice Data
  YM2413_PATCH  patch[19 * 2];
  long          patch_update[2]; // flag for check patch update
  unsigned long mask;
} extern YM2413;

// Create Object
void YM2413_Init(unsigned long, unsigned long);

// Setup
void YM2413_Reset(void);
void YM2413_ResetPatch(long);

void YM2413_SetRate(unsigned long);
void YM2413_SetQuality(unsigned long);
void YM2413_SetPan(unsigned long, unsigned long);

// Port/Register access
void YM2413_WriteIO(unsigned long reg, unsigned long val);
void YM2413_WriteReg(unsigned long reg, unsigned long val);

// Synthsize
void YM2413_Update(int **, int);
void YM2413_CalcStereo(long out[2]);

// Misc
void YM2413_SetPatch(const unsigned char *dump);
void YM2413_CopyPatch(long, YM2413_PATCH *);
void YM2413_ForceRefresh(void);
// Utility
void YM2413_Dump2Patch(const unsigned char *dump, YM2413_PATCH *patch);
void YM2413_Patch2Dump(const YM2413_PATCH *patch, unsigned char *dump);
void YM2413_GetDefaultPatch(long type, long num, YM2413_PATCH *);

// Channel Mask
unsigned long YM2413_SetMask(unsigned long mask);
unsigned long YM2413_ToggleMask(unsigned long mask);

#define dump2patch YM2413_dump2patch

#ifdef __cplusplus
}
#endif

#endif // _YM2413_H
